/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */
/* Adapted for Pharlap DOSX by G. Vanem 1997                 */
/* Adapted for Windows by G. Vanem 2004                      */

#include <stdio.h>
#include <signal.h>

#include "mwutil.h"

static int ungetch_x  = 0;

int char_avail = 0;

void mw_flush_kbd (void)
{
#if defined(WIN32)
  if (_mw_hStdin != INVALID_HANDLE_VALUE)
     FlushConsoleInputBuffer (_mw_hStdin);
#elif defined(MSDOS)
  /* !! to-do */
#endif
}

int getche (void)
{
  if (char_avail)
    /*
     * We don't know, whether the ungot char was already echoed.
     * We assume yes (for example in cscanf, probably the only
     * place where ungetch is ever called.
     * There is no way to check for this really, because
     * ungetch() could have been called with a character that
     * hasn't been got by a conio function. We don't echo again.
     */
     return (getch());
  return (putch(getch()));
}


int getch (void)
{
  int c;
#if defined(MSDOS)
  SWI_REGS regs;

  if (char_avail)
  {
    c = ungetch_x;
    char_avail = 0;
  }
  else
  {
    regs.eax = 0x0700;
    _dx_real_int (0x21, &regs);
    c = loBYTE (regs.eax);
  }
  return (c);

#elif defined(WIN32)
  int loop;

  if (char_avail)
  {
    c = ungetch_x;
    char_avail = 0;
    TRACE (("ungot '%c'/%d", c, c));
    return (c);
  }

  for (loop = 0; ; loop++)
  {
    INPUT_RECORD ir;
    const KEY_EVENT_RECORD *ker = &ir.Event.KeyEvent;
    DWORD num = 0;

    WaitForSingleObject (_mw_hStdin, INFINITE);
    if (!GetNumberOfConsoleInputEvents(_mw_hStdin, &num) || num == 0)
       continue;

    memset (&ir, 0, sizeof(ir));
    if (!ReadConsoleInput(_mw_hStdin, &ir, 1, &num))
    {
      TRACE (("ReadConsoleInput() failed; code %lu", GetLastError()));
      return (-1);
    }

    if (ir.EventType != KEY_EVENT)
       continue;

    if (/*!ker->dwControlKeyState && */ ker->bKeyDown)  /* don't want Ctrl keys */
    {
      TRACE (("%d: key down, ASCII %d, VK 0x%02X, kbhit_count %d (time %lu)",
              loop, ker->uChar.AsciiChar, ker->wVirtualKeyCode, _mw_kbhit_count,
              GetTickCount()));

      if (_mw_kbhit_count > 0)
          _mw_kbhit_count--;
      if (!ker->uChar.AsciiChar)    /* extended */
      {
        ungetch_x = ker->wVirtualKeyCode;
        char_avail = 1;
        return (0xE0);       /* double-sequence prefix */
      }
      return (ker->uChar.AsciiChar);
     }

    /* !!to-do: turn mouse wheel event into up/down arrow key-presses
     */
  }
#endif
}

int ungetch (int c)
{
  if (char_avail)
     return (EOF);
  ungetch_x  = c;
  char_avail = 1;
  return (c);
}

