#include <stdio.h>
#include <stdlib.h>
#include <io.h>

#include "mwutil.h"
#if defined(WIN32)
#error This is not for Windows. Ref. win32.c
#endif

int   _wscroll      = 1;
int   _near_screen  = 1;
int    conio_reinit = 0;
WORD   screen_seg   = 0xB800;
WORD   screen_start = 0;
WORD   screen_page  = 0;
WORD   CRT_base     = 0x3D4;
DWORD  linear_scr   = 0;

DWORD  line_ofs [MAX_SCR_HEIGHT];

BYTE _far *scr_fp;   /* far-pointer to screen */
BYTE      *scr_np;   /* near-pointer to screen (using memory mapping) */

void (*yield_proc)(void);

struct text_info _mw_ti = { 0,0, 80,25, 7,7, 3, 25,80, 0,0, 0 };

#ifdef __HIGHC__
  #pragma Offwarn(572)
#endif

#ifdef PHARLAP
  #define PEEKB(ofs)  PeekRealByte(ofs)
  #define PEEKW(ofs)  PeekRealWord(ofs)
#else  /* DOS4GW/X32VM extenders */
  #define PEEKB(ofs)  *(BYTE*)(ofs)
  #define PEEKW(ofs)  *(WORD*)(ofs)
#endif

#ifdef __WATCOMC__
  extern WORD My_DS(void);
  #pragma aux My_DS =  \
          "mov ax, ds" \
          modify [ax];

#elif defined(__DMC__)
  static WORD My_DS (void)
  {
    WORD _ax;
    __asm mov _ax, ds
    return (_ax);
  }
#endif

void mw_trace (const char *fmt, ...)
{
  (void) fmt;
}

static BYTE __getvideomode (void)
{
  BYTE mode = PEEKB (0x449);

  if (mode == C80 &&        /* = mode 3 */
      PEEKB(0x484) > 24)
     mode = C4350;
  return (mode);
}

static void __gettextinfo (void)
{
  BYTE x = wherex();
  BYTE y = wherey();
  BYTE c = scr_fp [GetVideoOfs(x,y) + 1];

  _mw_ti.currmode  = __getvideomode();
  _mw_ti.attribute = _mw_ti.normattr = c ? c : LIGHTGRAY;
}

void gettextinfo (struct text_info *ti)
{
  conio_init();
  *ti = _mw_ti;
}

static void WinYield (void)
{
  SWI_REGS regs;

  memset (&regs, 0, sizeof(regs));
  regs.eax = 0x1680;
  _dx_real_int (0x2F, &regs);
  if (loBYTE(regs.eax) == 0x80)   /* yield not supported by DPMI */
  {
    yield_proc = NULL;
    directcursor = 0;
  }
}

#if defined(__NNANSI)
  #define NNANSI_SLOW  "\x1B[?98l"
  #define NNANSI_FAST  "\x1B[?98h"
  #define ERASER       "\b\b\b\b\b\b      \r"

  static void nnansi_slow (void)
  {
    /* Put NNANSI.SYS in slow-write mode to prevent cursor
     * positioning trouble.
     */
    write (fileno(stdout), NNANSI_SLOW, sizeof(NNANSI_SLOW));
  }

  static void nnansi_fast (void)
  {
    BYTE x = wherex();

    write (fileno(stdout), NNANSI_FAST, sizeof(NNANSI_FAST));
    if (wherex() > x)
       write (fileno(stdout), ERASER, sizeof(ERASER));
  }
#endif  /* __NNANSI */

#ifdef PHARLAP
  static int init_scr_mapping (void)
  {
    DWORD mapping, pages, scr_size;

    scr_size = 2 * MAX_SCR_WIDTH * MAX_SCR_HEIGHT;
    pages    = (scr_size + 4095) & ~4096;

    if (_dx_map_phys(SS_DATA, linear_scr, pages, &mapping) == 0)
    {
      _mw_ti.mapping_scr = mapping;
      scr_np = (BYTE*)mapping;
      return (1);
    }
    _mw_ti.mapping_scr = 0UL;
    _near_screen = FALSE;
    scr_np = NULL;
    return (0);
  }
#endif

void conio_init (void)
{
  static char is_init = 0;

  if (!is_init || conio_reinit)
  {
    SWI_REGS regs;
    BYTE     w, h, y, videoMode;

#if !defined(__WATCOMC__)  /* WASM is so buggy! */
    if (tinfo_size != sizeof(_mw_ti))
    {
      printf ("%s(%u): Development error: size mismatch:\n"
              "  tinfo_size=%lu vs sizeof(_mw_ti)=%u\n",
              __FILE__, __LINE__, tinfo_size, sizeof(_mw_ti));
      exit (-1);
    }
#endif
    if (3 & (unsigned)&_mw_ti.mapping_scr)
    {
      printf ("%s(%u): Development error: mapping_scr should be "
              "DWORD aligned\n", __FILE__, __LINE__);
      exit (-1);
    }

    CRT_base = PEEKW (0x463);
    regs.eax = 0x0F00;
    _dx_real_int (0x10, &regs);
    videoMode    = loBYTE (regs.eax);
    screen_seg   = videoMode == MONO ? 0xB000 : 0xB800;
    screen_page  = hiBYTE (regs.ebx);
    screen_start = 0x1000 * screen_page;

    _mw_ti.screenwidth  = w = hiBYTE (regs.eax);
    _mw_ti.screenheight = h = 1 + PEEKB (0x484);
    _mw_ti.wintop       = 1;
    _mw_ti.winleft      = 1;
    _mw_ti.winright     = w;
    _mw_ti.winbottom    = h;

    linear_scr = screen_seg << 4;

    for (y = 0; y < MAX_SCR_HEIGHT; y++)
        line_ofs[y] = screen_start + ((y * w) << 1);

#ifdef PHARLAP
    if (_near_screen)
       init_scr_mapping();
#endif

    if (!is_init)
    {
      is_init = 1;
      yield_proc = WinYield;
      WinYield();

#ifdef __NNANSI  /* Not worth it. nnansi_fast() may mess with the screen */
      nnansi_slow();
      atexit (nnansi_fast);
#endif
    }
    conio_reinit = 0;

#ifdef PHARLAP
    FP_SET (scr_fp, linear_scr, SS_DOSMEM);
#else
    scr_fp = MK_FP (My_DS(), linear_scr);
    scr_np = (BYTE*) linear_scr;
#endif
    __gettextinfo();
  }
}

/*-----------------------------------------------------------------------*/

#if (defined(__WATCOMC__) && !defined(PHARLAP)) || \
    (defined(__DMC__) && defined(DOS386))
void _dx_real_int (int inum, SWI_REGS *reg)
{
  union  REGS  r;
  struct SREGS s;

  memset (&r, 0, sizeof(r));
  segread (&s);

#if defined(__DMC__)
  r.x.ax  = 0x300;
  r.e.ebx = inum;
  r.x.cx  = 0;
  s.es    = FP_SEG (reg);
  r.e.edi = FP_OFF (reg);
  reg->flags = 0;
  reg->ss = reg->sp = 0;
  int86x_real (0x31, &r, &r, &s);
#else
  r.w.ax  = 0x300;
  r.x.ebx = inum;
  r.w.cx  = 0;
  s.es    = FP_SEG (reg);
  r.x.edi = FP_OFF (reg);
  reg->flags = 0;
  reg->ss = reg->sp = 0;
  int386x (0x31, &r, &r, &s);
#endif
}
#endif
