#include <stdio.h>
#include <stdlib.h>
#include "mwutil.h"

#ifdef __HIGHC__
extern unsigned _mwstack_limit[2];
#endif

/*
 * (x1,y1) is physical upper lev corner,
 * (x2,y2) is physical lower right corner to write 'buf' to.
 * 'buf' is a sequence of 8-bits alternating between ascii
 * value and character attribute.
 *
 * For Win32, we must clone this into a new array of CHAR_INFO.
 */
int puttext (int x1, int y1, int x2, int y2, const void *buf)
{
#if defined(MSDOS)
  register DWORD xy;
  register int   wd = (x2-x1+1) << 1;
#endif
  register char *p  = (char*) buf;

  conio_init();

#if defined(MSDOS)
  xy = GetVideoOfs (x1-1, y1-1);

  for (; y1 <= y2; y1++)
  {
    SCREEN_WRITE (xy, p, wd);
    xy += 2 * _mw_ti.screenwidth;
    p  += wd;
  }
#elif defined(WIN32)
  {
    SMALL_RECT rect;
    COORD      coord, ci_size;
    CHAR_INFO *ci, *ci_start;
    int        i, size;

    size = (x2-x1+1) * (y2-y1+1);
    ci = alloca (sizeof(*ci) * size);
    if (!ci)
       return (-1);

    ci_start = ci;
    for (i = 0; i < size; i++, ci++)
    {
      ci->Char.AsciiChar = *p++;
      ci->Attributes     = *p++;
    }
    ci_size.X = x2 - x1 + 1;
    ci_size.Y = y2 - y1 + 1;
    coord.X   = 0;
    coord.Y   = 0;
    rect.Left   = x1 - 1;
    rect.Right  = x2 - 1;
    rect.Top    = y1 - 1 + _mw_csbi.srWindow.Top;
    rect.Bottom = y2 - 1 + _mw_csbi.srWindow.Top;
    WriteConsoleOutput (_mw_hStdout, ci_start, ci_size, coord, &rect);
  }
#endif
  return (1);
}

#undef gettext  /* A built-in in MingW */
int gettext (int x1, int y1, int x2, int y2, void *buf)
{
  return mw_gettext (x1, y1, x2, y2, buf);
}

int mw_gettext (int x1, int y1, int x2, int y2, void *buf)
{
#if defined(MSDOS)
  register DWORD xy;
  register int   wd = (x2-x1+1) << 1;
#endif
  register char *p  = (char*) buf;

  conio_init();

#if defined(MSDOS)
  xy = GetVideoOfs (x1-1, y1-1);

  for (; y1 <= y2; y1++)
  {
    SCREEN_READ (p, xy, wd);
    xy += 2 * _mw_ti.screenwidth;
    p  += wd;
  }
#elif defined(WIN32)
  {
    SMALL_RECT rect;
    COORD      coord, ci_size;
    CHAR_INFO *ci, *ci_start;
    int        i, size;

    size = (x2-x1+1) * (y2-y1+1);
    ci = alloca (sizeof(*ci) * size);
    if (!ci)
       return (-1);

    ci_start  = ci;
    ci_size.X = x2 - x1 + 1;
    ci_size.Y = y2 - y1 + 1;
    coord.X   = 0;
    coord.Y   = 0;
    rect.Top    = y1 - 1 + _mw_csbi.srWindow.Top;
    rect.Bottom = y2 - 1 + _mw_csbi.srWindow.Top;
    rect.Left   = x1 - 1;
    rect.Right  = x2 - 1;
    ReadConsoleOutput (_mw_hStdout, ci_start, ci_size, coord, &rect);
    for (i = 0; i < size; i++, ci++)
    {
      *p++ = ci->Char.AsciiChar;
      *p++ = ci->Attributes;
    }
  }
#endif
  return (1);
}

int movetext (int x1, int y1, int x2, int y2, int to_x, int to_y)
{
  int   size = (x2-x1+1) * (y2-y1+1) << 1;
  char *buf;

#ifdef __HIGHC__
  if ((unsigned)&buf < size + _mwstack_limit[1])
  {
    fprintf (stderr, "FATAL: stack too small in movetext() (%u needed)\n",
             size);
    return (0);
  }
#endif

  buf = alloca (size);
  if (!buf)
     return (0);

  mw_gettext (x1, y1, x2, y2, buf);
  puttext (to_x, to_y, to_x+x2-x1, to_y+y2-y1, buf);
  return (1);
}

