#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <malloc.h>
#include <tchar.h>

#include "mwutil.h"

int directvideo  = 1;
int directcursor = 0;
int dpmi_yield   = 0;
int _wscroll     = 1;
int _near_screen = 0;

BOOL   conio_reinit = 0;
HANDLE _mw_hStdin   = INVALID_HANDLE_VALUE;
HANDLE _mw_hStdout  = INVALID_HANDLE_VALUE;
HANDLE _mw_hStderr  =  INVALID_HANDLE_VALUE;

struct text_info _mw_ti = { 0, 0, 80, 25, 7, 7, 3, 25, 80, 0, 0, 0, 0 };

FILE *_mw_debug_file;
int   _mw_debug_level;
int   _mw_kbhit_count = 0;

CONSOLE_SCREEN_BUFFER_INFO _mw_csbi;

static void setup_console (void)
{
  DWORD mode;

  if (GetConsoleMode(_mw_hStdin, &mode))
  {
    mode &= ~(ENABLE_ECHO_INPUT | ENABLE_LINE_INPUT |
              ENABLE_PROCESSED_INPUT | ENABLE_WINDOW_INPUT);
    SetConsoleMode (_mw_hStdin, mode);
  }
}

static void setup_mouse (BOOL on)
{
  DWORD mode;

  if (GetConsoleMode(_mw_hStdin, &mode))
  {
    if (on)
         mode |= ENABLE_MOUSE_INPUT;
    else mode &= ~ENABLE_MOUSE_INPUT;
    SetConsoleMode (_mw_hStdin, mode);
  }
}

void gettextinfo (struct text_info *ti)
{
  conio_init();
  *ti = _mw_ti;
}

void conio_init (void)
{
  static BOOL is_init = FALSE;

  if (!is_init || conio_reinit)
  {
    DWORD       read;
    WORD        attr;
    const char *env = getenv ("MWCONIO");

    if (env)
    {
      _mw_debug_level = atoi (env);
      if (_mw_debug_level > 0)
      {
        const char *p = strchr (env,',');

        if (p && p[1])
             _mw_debug_file = fopen (p+1, "w+t");
        else _mw_debug_file = stderr;
      }
    }

    _mw_hStdin  = GetStdHandle (STD_INPUT_HANDLE);
    _mw_hStdout = GetStdHandle (STD_OUTPUT_HANDLE);
    _mw_hStderr = GetStdHandle (STD_ERROR_HANDLE);

    if (_mw_hStdin  == INVALID_HANDLE_VALUE ||
        _mw_hStdout == INVALID_HANDLE_VALUE ||
        _mw_hStderr == INVALID_HANDLE_VALUE)
    {
      TRACE (("Bad handles; hStdin %lu, hStdout %lu, hStderr %lu",
              (DWORD)_mw_hStdin, (DWORD)_mw_hStdout, (DWORD)_mw_hStderr));
      return;
    }

    is_init = TRUE;
    conio_reinit = FALSE;
    memset (&_mw_csbi, 0, sizeof(_mw_csbi));
    if (!GetConsoleScreenBufferInfo(_mw_hStdout, &_mw_csbi))
    {
      TRACE (("GetConsoleScreenBufferInfo() failed; code %lu",
              GetLastError()));
      return;
    }

    setup_console();
    setup_mouse (FALSE);

    _mw_ti.wintop       = 1;
    _mw_ti.winleft      = 1;
    _mw_ti.screenwidth  = _mw_csbi.srWindow.Right - _mw_csbi.srWindow.Left + 1;
    _mw_ti.screenheight = _mw_csbi.srWindow.Bottom - _mw_csbi.srWindow.Top + 1;
    _mw_ti.winright     = _mw_ti.screenwidth;
    _mw_ti.winbottom    = _mw_ti.screenheight;
    _mw_ti.curx         = _mw_csbi.dwCursorPosition.X;
    _mw_ti.cury         = _mw_csbi.dwCursorPosition.Y - _mw_csbi.srWindow.Top + 1;
    _mw_ti.scrollwidth  = _mw_csbi.dwSize.X;
    _mw_ti.scrollheight = _mw_csbi.dwSize.Y;
    ReadConsoleOutputAttribute (_mw_hStdout, &attr, 1, _mw_csbi.dwCursorPosition, &read);
    _mw_ti.attribute = attr;
    _mw_ti.normattr  = attr;
    TRACE (("winright %lu, winbottom %lu, curx %lu, cury %lu, max-x %d, max-y %d",
            _mw_ti.winright, _mw_ti.winbottom,
            _mw_ti.curx, _mw_ti.cury,
            _mw_csbi.dwMaximumWindowSize.X, _mw_csbi.dwMaximumWindowSize.Y));
  }
}

void mw_trace (const char *fmt, ...)
{
  va_list arg;

  if (!_mw_debug_file || _mw_debug_level <= 0)
     return;

  va_start (arg, fmt);
  vfprintf (_mw_debug_file, fmt, arg);
  va_end (arg);
}

int mw_kbhit (void)
{
  INPUT_RECORD *ir;
  DWORD num, written, i;

  if (_mw_kbhit_count > 0)
     return (1);

  if (!GetNumberOfConsoleInputEvents(_mw_hStdin, &num) || num == 0)
  {
    Sleep (0);
    return (0);
  }

  ir = alloca (num * sizeof(*ir));
  if (!ir)
     return (0);

  memset (ir, 0, num * sizeof(*ir));
  if (!ReadConsoleInput(_mw_hStdin, ir, num, &num))
  {
    TRACE (("ReadConsoleInput() failed; code %lu", GetLastError()));
    return (0);
  }

  for (i = 0; i < num; i++, ir++)
  {
    const KEY_EVENT_RECORD   *ker = &ir->Event.KeyEvent;
    const MOUSE_EVENT_RECORD *mer = &ir->Event.MouseEvent;

    switch (ir->EventType)
    {
      case KEY_EVENT:
           if (ker->bKeyDown)
              TRACE (("%lu: key down, ASCII %d, VK 0x%02X, state 0x%02lX, kbhit_count %d (time %lu)",
                      i, ker->uChar.AsciiChar, ker->wVirtualKeyCode,
                      ker->dwControlKeyState, _mw_kbhit_count, GetTickCount()));

           if (ker->bKeyDown /* && !ker->dwControlKeyState */)
           {
             /* put the record back */
             WriteConsoleInput (_mw_hStdin, ir, 1, &written);
             _mw_kbhit_count++;
             return (1);
           }
           break;

      case MOUSE_EVENT:
           TRACE (("%lu: mouse-event", i));
#if 0
           // push-back
           if (mer->dwEventFlags & MOUSE_WHEELED)
           {
	     const char *dir = (mer->dwButtonState & 0xFF000000) ? "down" : "up";
             // push-back 4 up/down arrow key codes
             _mw_kbhit_count += 4;
             return (1);
           }
#endif
           break;
    }
  }
  return (0);
}

void gotoxy_abs (int x, int y)
{
  COORD coord;

  coord.X = x;
  coord.Y = y + _mw_csbi.srWindow.Top;  /* place in display window */
  SetConsoleCursorPosition (_mw_hStdout, coord);
  TRACE (("x %d, y %d", coord.X, coord.Y));
}

WORD wherexy_abs (void)
{
  WORD x, y;

  GetConsoleScreenBufferInfo (_mw_hStdout, &_mw_csbi);
  x = _mw_csbi.dwCursorPosition.X;
  y = _mw_csbi.dwCursorPosition.Y - _mw_csbi.srWindow.Top;  /* relative display window */
  if (x > 255 || y > 255)
       TRACE (("x %d, y %d !! type overflow", x, y));
  else TRACE (("x %d, y %d", x, y));
  return (x + (y << 8));
}

void gotoxy (int x, int y)
{
  conio_init();
  _mw_ti.curx = x;
  _mw_ti.cury = y;
  gotoxy_abs (x+_mw_ti.winleft-2, y+_mw_ti.wintop-2);
}

int wherex (void)
{
  WORD xy;

  conio_init();
  xy = wherexy_abs();
  _mw_ti.curx = loBYTE(xy) - _mw_ti.winleft + 2;
  return (_mw_ti.curx);
}

int wherey (void)
{
  WORD xy;

  conio_init();
  xy = wherexy_abs();
  _mw_ti.cury = hiBYTE(xy) - _mw_ti.wintop + 2;
  return (_mw_ti.cury);
}

void textmode (int mode)
{
  TRACE (("textmode not implemented"));
  (void) mode;
}

void sound (unsigned int freq)
{
  (void) freq;
}

DWORD FastWrite (int x, int y, int col, const char *str)
{
  WORD  attr;
  TCHAR ch[1];
  COORD coord;
  DWORD written;

  while ((ch[0] = *str) != '\0')
  {
    switch (ch[0])
    {
      case '\a':
           MessageBeep (MB_ICONEXCLAMATION);
           break;

      case '\n':
           ++y;
           break;

      case '\r':
           x = _mw_ti.winleft - 1;
           break;

      case '\b':
           if (x > (int)_mw_ti.winleft-1)
               x--;
           else if (y > (int)_mw_ti.wintop-1)
           {
             y--; /* Creeps up one line. Should we do this? */
             x = _mw_ti.winright - 1;
           }
           break;

      default:
           attr = col;
           coord.X = x;
           coord.Y = y + _mw_csbi.srWindow.Top;
           WriteConsoleOutputCharacter (_mw_hStdout, ch, 1, coord, &written);
           WriteConsoleOutputAttribute (_mw_hStdout, &attr, 1, coord, &written);
           x++;
           break;
    }

    if (x > (int)_mw_ti.winright-1)
    {
      y++;
      x = _mw_ti.winleft - 1;
    }
    if (y > (int)_mw_ti.winbottom-1)
    {
      SCROLL_UP (_mw_ti.winleft, _mw_ti.wintop,
                 _mw_ti.winright,_mw_ti.winbottom, col);
      y--;
    }
    str++;
  }
  return (x + y * _mw_ti.screenwidth);
}
